import pickle
from matplotlib import pyplot as plt
import seaborn as sns

from numpy import random
import numpy as np




'''SCL and FSN data:'''

S_scl = pickle.load(open('rank2length24iter50000.p','rb'))
S_scl2 = [2*s for s in S_scl]
S_fsn = pickle.load(open('random_scl_graphsize25iter50000.p','rb'))




'''The X distribution'''

class X_dist():
    '''Computes an array of length N of the X distribution. Default is the X distribution which models scl.'''
    def __init__(self, d = -2, beta = -0.2, mu = 2.164, c1 = 0.3, c2 = -0.14, N = 50000):
        self.d = d
        self.beta = beta
        self.mu = mu
        self.c1 = c1
        self.c2 = c2
        self.N = N
        self.X = [0]*N
        self.denominators = [1]*N



    def compute_denominators(self):
        '''computes the denominator distribution of X. This is the distribution on ...
        the integers >= 1 with pmf proportional to n**(d*(1-n**beta))'''

        total_sum = 0
        for i in range(1, 5000):
            total_sum += self.pmf_denominators(i)

        vec = []
        for i in range(self.N):
            u = random.uniform(0, 1)
            n = 1
            sum = self.pmf_denominators(1)
            total = u * total_sum
            while sum <= total:
                n += 1
                sum += self.pmf_denominators(n)

            vec.append(n)
        self.denominators = vec


    def pmf_denominators(self, n):
        return n ** (self.d * (1 - n ** self.beta))



    def compute(self):
        vec = []
        self.compute_denominators()
        for i in range(self.N):
            n = self.denominators[i]
            a = random.normal(self.mu, self.c1 * (n ** (self.c2)))
            res = float(round(a * n) / n)
            vec.append(res)

        self.X = vec
















class analyse_distributions():

    def denominator(self, n, S):
        '''returns all the elements in Set with denominator exactly n.'''
        divisors = []
        round_to = 7
        for i in range(1, n):
            if n % i == 0:
                divisors.append(i)
        S_denominators = []
        for s in S:
            if round(n * s, round_to) % 1 == 0:
                proper_divisor = True
                for d in divisors:
                    if round(d * s, round_to) % 1 == 0:
                        proper_divisor = False

                if proper_divisor:
                    s = round(s, round_to)
                    S_denominators.append(s)

        return S_denominators

    def find_denominator(self, S):
        '''Given a list of rationals S it returns a list of denominators for every entry of S'''
        tol = 10 ** (-6)
        D = [1] * len(S)
        for i in range(len(S)):
            for j in range(1, 1000):
                s = S[i]
                if abs(round(j * s) - j * s) <= tol:
                    D[i] = j
                    break

        return D

    def denominators_totient_clean(self, S):
        '''for a list of rationals S it returns denominators n and adds n/phi(n) of the values of n,
        where phi is Euler's totient function.'''
        D = self.find_denominator(S)
        D_new = []
        for d in D:
            D_new.append(d)
            sum = random.uniform(0, 1)
            while d / self.phi(d) > sum:
                D_new.append(d)
                sum += random.uniform(0, 1)

        return D_new


    '''a bit of number theory .. '''

    def gcd(self, a, b):
        if (self, a == 0):
            return b
        return gcd(b % a, a)

    def phi(self, n):
        '''Euler's totient function'''
        result = 1
        for i in range(2, n):
            if (self.gcd(i, n) == 1):
                result += 1
        return result

    def is_prime(self, a):
        x = True
        for i in range(2, a):
            if a % i == 0:
                x = False

        return x





    '''Estimations'''

    def fit_exponential(self, X, Y):
        '''function to determine C and theta in y = C*x^theta. Here X and Y are arrays
        which satisfy this equation and we estimate C and theta using the LSE'''
        X_log = np.log(X)
        Y_log = np.log(Y)

        [theta, c] = np.polyfit(X_log, Y_log, 1)
        return theta, np.exp(c)


    def different_variances(self, S, limit):
        '''computes theta, C, where we suppose that the standard deviation of all
        denominators with n is C*n^(theta)'''
        Primes = []
        Variances = []
        for i in range(2, limit):
            if i == 1 or self.is_prime(i):
                S_n = self.denominator(i, S)
                Primes.append(i)
                Variances.append(np.var(S_n))

        return self.fit_exponential(Primes, np.sqrt(Variances))

    def different_denominators(self, S, limit):
        '''If frequency of elements in S is n**theta, it returns theta.'''
        Primes = []
        Length = []
        for i in range(7, limit):
            if self.is_prime(i):
                S_n = self.denominator(i, S)
                Primes.append(i)
                Length.append(len(S_n))

        # plt.scatter(np.log(Primes), np.log(Length))
        return self.fit_exponential(Primes, Length)

    def Estimate_Parameters_X(self, S, depth):
        '''If S is an X distribution, it returns an estimation of the parameters of X'''
        mu = np.mean(S)
        d, x = self.different_denominators(S, depth)
        c_2, c_1 = self.different_variances(S, depth)
        print('Parameters: ' + ' d = ' + str(d) + ', mu = ' + str(mu) + ', c_1 = ' + str(c_1) + ', c_2 = ' + str(
            c_2) + '.')






    '''histograms and plots'''

    def histogram_denominators(self, S, name, name_label):
        '''plots a logarithmic histogram of the denominators in S'''
        D = self.find_denominator(S)
        plt.figure(name)
        plt.yscale('log')
        plt.xscale('log')
        plt.hist(D, bins=3000, label=name_label)
        plt.legend(loc='upper right')

    def histogram_denominators_totient_clean(self, S, name, name_label):
        '''plots a logarithmic histogram of the denominators in S where spikes have been corrected by Euler's Totient function.'''
        D = self.denominators_totient_clean(S)
        plt.figure(name)
        plt.yscale('log')
        plt.xscale('log')
        plt.hist(D, bins=3000, label=name_label)
        plt.legend(loc='upper right')





    def plot_different_denominators_fixed_denominator(self, S, l):
        '''plots histograms of all values which have denominators in l'''
        for n in l:
            S_n = self.denominator(n, S)
            plt.hist(S_n, bins=400, label='n = ' + str(n))
            plt.legend(loc='upper right')

        plt.show()


    def plot_different_denominators(self, S, limit, name):
        '''plots the different frequencies of elements with
        prime denominators in S'''
        Primes = []
        Length = []
        for i in range(1, limit):
            if i == 1 or self.is_prime(i):
                S_n = self.denominator(i, S)
                Primes.append(i)
                Length.append(len(S_n) * self.phi(i) / i)

        plt.scatter(Primes, Length, label=name)
        plt.yscale('log')
        plt.xscale('log')
        plt.xlabel('denominator n')
        plt.ylabel('$\phi(n)/n$ * e_n')

    def plot_different_std(self, S, limit, name):
        Primes = []
        Length = []
        for i in range(1, limit):
            if i == 1 or self.is_prime(i):
                S_n = self.denominator(i, S)
                Primes.append(i)
                Length.append(np.sqrt(np.var(S_n)))


        plt.scatter(Primes, Length, label=name)
        plt.yscale('log')
        plt.xscale('log')
        plt.xlabel('denominator n')


    def compare_histograms(self, S1, S2, lower, upper, bins, S1_label, S2_label, denominator, largest_frequency):

        x_pos, x_labels = create_x_ticks(lower, upper, denominator, 10)



        plt.figure(figsize=(25, 4), frameon=True)
        plt.hist([S1, S2], bins=np.linspace(lower, upper, bins), label=[S1_label, S2_label])

        plt.xticks(x_pos, x_labels)
        plt.legend(loc='upper right')

        axes = plt.gca()
        axes.set_ylim([0, largest_frequency])
        plt.show()



def create_x_ticks(lower, upper, denominator, n_limit):
    x_pos = []
    x_labels = []
    for d in range(1, denominator+1):
        for i in range(n_limit*d):
            if not i/d in x_pos:
                if lower <= i/d <= upper:
                    x_pos.append(i/d)
                    if d == 1:
                        x_labels.append(str(i))
                    else:
                        x_labels.append('$\dfrac{' + str(i)+ '}{' + str(d) + '}$')

    return x_pos, x_labels









'''Functions to Create the Figures:'''


def histogram_scl_plain():
    x_pos, x_labels = create_x_ticks(0.8, 1.4, 9, 4)
    S_scl_clean = [round(s, 4) for s in S_scl]


    '''histogram'''
    bins = np.linspace(0.8,1.4,1000)
    plt.figure(figsize=(25,4), frameon=True)
    plt.hist(S_scl_clean, bins = bins)
    plt.xticks(x_pos, x_labels)
    axes = plt.gca()
    axes.set_ylim([0,1300])


    plt.show()



def histogram_half_fsn_plain():
    x_pos, x_labels = create_x_ticks(2.5, 3.6, 7, 4)
    S_fsn_clean = [0.5*round(s, 4) for s in S_fsn]


    '''histogram'''
    bins = np.linspace(2.5,3.6,1000)
    plt.figure(figsize=(25,4), frameon=True)
    plt.hist(S_fsn_clean, bins = bins)
    plt.xticks(x_pos, x_labels)
    axes = plt.gca()
    axes.set_ylim([0,2000])


    plt.show()



def X_vs_scl():
    A = analyse_distributions()
    X = X_dist()
    X.compute()
    S = X.X

    A.compare_histograms(S_scl2, S, 1.4, 2.8, 500, 'SCL', 'X', 7, 3000)



def X_vs_fsn():
    A = analyse_distributions()
    X = X_dist(d = -2.8, beta = -0.2, mu = 6.141, c1 = 0.5, c2 = -0.1, N = 50000)
    X.compute()
    S = X.X

    A.compare_histograms(S_fsn, S, 5.3, 6.8, 500, 'FSN', 'X', 5, 2000)




#X_vs_fsn()
#X_vs_scl()
histogram_half_fsn_plain()
#histogram_scl_plain()



'''Denominator Power Picture:
A.plot_different_denominators(S_fsn, limit=100, name='FSN')
A.plot_different_denominators(S_scl2, limit=100, name='SCL')
plt.legend()
plt.show()
'''


'''Different denominator distribution picture
A.plot_different_denominators_fixed_denominator(S_scl2, [13, 23, 29])
'''

'''Denominator Standard Deviation Picture:
A.plot_different_std(S_fsn, limit=100, name='FSN')
A.plot_different_std(S_scl2, limit=100, name='SCL')
plt.legend()
plt.show()
'''

'''Estimating the parameters of SCL
A.Estimate_Parameters_X(S_scl2, 40)
'''


'''Generating an X distribution for SCL
X = X_dist()
X.compute()
S = X.X
'''

'''Plot the histograms
A.compare_histograms(S_scl2, S, 1.5, 2.5, 500, 'SCL', 'X')
'''




